from cms.app_base import CMSApp
from cms.apphook_pool import apphook_pool
from django.conf.urls import url
from django.db import transaction
from django.utils.translation import ugettext_lazy as _

from . import views
from .urls import LEPRIKON_URL


@apphook_pool.register
class LeprikonApp(CMSApp):
    name = _('Leprikon')
    app_name = 'leprikon'

    def get_urls(self, page=None, language=None, **kwargs):
        return ['leprikon.urls']


@apphook_pool.register
class LeprikonSubjectTypeApp(CMSApp):
    name = _('Leprikon subject type')
    app_name = 'leprikon_subject_type'

    def get_urls(self, page=None, language=None, **kwargs):
        return [
            url(r'^$', views.subject_list, name='subject_list'),
            url(r'^(?P<pk>[0-9]+)/$', views.subject_detail, name='subject_detail'),
            url(r'^(?P<pk>[0-9]+)/{registration}/$'.format(**LEPRIKON_URL),
                transaction.atomic(views.subject_registration_form),
                name='subject_registration_form'),
        ]



from datetime import date
from json import dumps

from django import forms
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import transaction
from django.db.models import Q
from django.forms.models import inlineformset_factory
from django.utils.functional import cached_property
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _

from ..models.agegroup import AgeGroup
from ..models.courses import Course, CourseRegistration
from ..models.department import Department
from ..models.events import Event, EventRegistration
from ..models.fields import DAY_OF_WEEK
from ..models.leprikonsite import LeprikonSite
from ..models.place import Place
from ..models.roles import BillingInfo, Leader, Parent, Participant
from ..models.school import School
from ..models.subjects import (
    Subject, SubjectGroup, SubjectRegistration, SubjectRegistrationBillingInfo,
    SubjectRegistrationGroup, SubjectRegistrationGroupMember,
    SubjectRegistrationParticipant, SubjectType,
)
from ..utils import get_age, get_birth_date
from .fields import AgreementBooleanField
from .form import FormMixin
from .widgets import RadioSelectBootstrap

User = get_user_model()


class SubjectFilterForm(FormMixin, forms.Form):
    q = forms.CharField(label=_('Search term'), required=False)
    course_types = forms.ModelMultipleChoiceField(queryset=None, label=_('Course type'), required=False)
    event_types = forms.ModelMultipleChoiceField(queryset=None, label=_('Event type'), required=False)
    departments = forms.ModelMultipleChoiceField(queryset=None, label=_('Department'), required=False)
    groups = forms.ModelMultipleChoiceField(queryset=None, label=_('Group'), required=False)
    leaders = forms.ModelMultipleChoiceField(queryset=None, label=_('Leader'), required=False)
    places = forms.ModelMultipleChoiceField(queryset=None, label=_('Place'), required=False)
    age_groups = forms.ModelMultipleChoiceField(queryset=None, label=_('Age group'), required=False)
    days_of_week = forms.MultipleChoiceField(label=_('Day of week'),
                                             choices=tuple(sorted(DAY_OF_WEEK.items())), required=False)
    past = forms.BooleanField(label=_('Include past subjects'), required=False)
    reg_active = forms.BooleanField(label=_('Available for registration'), required=False)
    invisible = forms.BooleanField(label=_('Show invisible'), required=False)

    _models = {
        SubjectType.COURSE: Course,
        SubjectType.EVENT: Event,
    }

    def __init__(self, subject_type_type, subject_types, school_year, is_staff, data, **kwargs):
        super(SubjectFilterForm, self).__init__(data=data, **kwargs)
        self.subject_type_type = subject_type_type

        # pre filter subjects by initial params
        qs = self._models[subject_type_type].objects.filter(school_year=school_year)
        if len(subject_types) == 1:
            qs = qs.filter(subject_type=subject_types[0])
        else:
            qs = qs.filter(subject_type__in=subject_types)
        if not is_staff or 'invisible' not in data:
            qs = qs.filter(public=True)
        self.qs = qs

        subject_ids = tuple(qs.order_by('id').values_list('id', flat=True).distinct())
        if len(subject_types) == 1:
            del self.fields['course_types']
            del self.fields['event_types']
        elif subject_type_type == SubjectType.COURSE:
            del self.fields['event_types']
            self.fields['course_types'].queryset = SubjectType.objects.filter(id__in=(st.id for st in subject_types))
        elif subject_type_type == SubjectType.EVENT:
            del self.fields['course_types']
            self.fields['event_types'].queryset = SubjectType.objects.filter(id__in=(st.id for st in subject_types))

        self.fields['departments'].queryset = Department.objects.filter(subjects__id__in=subject_ids).distinct()
        if self.fields['departments'].queryset.count() == 0:
            del self.fields['departments']

        self.fields['groups'].queryset = SubjectGroup.objects.filter(subject_types__in=subject_types,
                                                                     subjects__id__in=subject_ids).distinct()
        if self.fields['groups'].queryset.count() == 0:
            del self.fields['groups']

        self.fields['leaders'].queryset = (Leader.objects.filter(subjects__id__in=subject_ids).distinct()
                                           .order_by('user__first_name', 'user__last_name'))
        if self.fields['leaders'].queryset.count() == 0:
            del self.fields['leaders']

        self.fields['places'].queryset = Place.objects.filter(subjects__id__in=subject_ids).distinct()
        if self.fields['places'].queryset.count() == 0:
            del self.fields['places']

        self.fields['age_groups'].queryset = AgeGroup.objects.filter(subjects__id__in=subject_ids).distinct()
        if self.fields['age_groups'].queryset.count() <= 1:
            del self.fields['age_groups']

        if subject_type_type != SubjectType.COURSE:
            del self.fields['days_of_week']
        if subject_type_type != SubjectType.EVENT:
            del self.fields['past']
        if not is_staff:
            del self.fields['invisible']

        for f in self.fields:
            self.fields[f].help_text = None

    def get_queryset(self):
        if not self.is_valid():
            return self.qs
        qs = self.qs
        for word in self.cleaned_data['q'].split():
            qs = qs.filter(
                Q(name__icontains=word) |
                Q(description__icontains=word)
            )
        if self.cleaned_data.get('course_types'):
            qs = qs.filter(subject_type__in=self.cleaned_data['course_types'])
        elif self.cleaned_data.get('event_types'):
            qs = qs.filter(subject_type__in=self.cleaned_data['event_types'])
        if self.cleaned_data.get('departments'):
            qs = qs.filter(department__in=self.cleaned_data['departments'])
        if self.cleaned_data.get('groups'):
            qs = qs.filter(groups__in=self.cleaned_data['groups'])
        if self.cleaned_data.get('places'):
            qs = qs.filter(place__in=self.cleaned_data['places'])
        if self.cleaned_data.get('leaders'):
            qs = qs.filter(leaders__in=self.cleaned_data['leaders'])
        if self.cleaned_data.get('age_groups'):
            qs = qs.filter(age_groups__in=self.cleaned_data['age_groups'])
        if self.cleaned_data.get('days_of_week'):
            qs = qs.filter(times__day_of_week__in=self.cleaned_data['days_of_week'])
        if self.subject_type_type == SubjectType.EVENT and not self.cleaned_data['past']:
            qs = qs.filter(end_date__gte=now())
        if self.cleaned_data['reg_active']:
            qs = qs.filter(reg_from__lte=now()).exclude(reg_to__lte=now()).exclude(price=None)
        return qs.distinct()


class SubjectForm(FormMixin, forms.ModelForm):

    class Meta:
        model = Subject
        fields = ['description', 'risks', 'plan', 'evaluation']


class SubjectAdminForm(FormMixin, forms.ModelForm):

    def __init__(self, data=None, *args, **kwargs):
        super().__init__(data, *args, **kwargs)
        instance = kwargs.get('instance')
        leprikon_site = LeprikonSite.objects.get_current()

        # limit choices of groups
        groups_choices = self.fields['groups'].widget.choices
        groups_choices.queryset = groups_choices.queryset.filter(
            subject_types__subject_type=self.subject_type
        ).distinct()

        # limit choices of leaders
        leaders_choices = self.fields['leaders'].widget.choices
        leaders_choices.queryset = leaders_choices.queryset.filter(school_years=self.school_year)

        # limit choices of questions
        if instance:
            questions_choices = self.fields['questions'].widget.choices
            questions_choices.queryset = questions_choices.queryset.exclude(
                id__in=instance.subject_type.questions.values('id')
            )

        # limit choices of registration agreements
        registration_agreements_choices = self.fields['registration_agreements'].widget.choices
        registration_agreements_choices.queryset = registration_agreements_choices.queryset.exclude(
            id__in=leprikon_site.registration_agreements.values('id')
        )
        if instance:
            registration_agreements_choices.queryset = registration_agreements_choices.queryset.exclude(
                id__in=instance.subject_type.registration_agreements.values('id')
            )


class SchoolMixin:
    def _add_error_required(self, field_name):
        self.add_error(field_name, forms.ValidationError(
            self.fields[field_name].error_messages['required'],
            code='required',
        ))

    def clean(self):
        # check required school and class
        x_group = self.cleaned_data.get(self.x_group)
        if x_group and x_group.require_school:
            # require school
            school = self.cleaned_data.get('school')
            if school:
                if school == 'other':
                    self.cleaned_data['school'] = None
                    # require other school
                    if not self.cleaned_data.get('school_other'):
                        self._add_error_required('school_other')
                else:
                    self.cleaned_data['school'] = School.objects.get(
                        id=int(self.cleaned_data['school']),
                    )
            else:
                self._add_error_required('school')
            # require school class
            if not self.cleaned_data.get('school_class'):
                self._add_error_required('school_class')
        else:
            # delete school and class
            self.cleaned_data['school'] = None
            self.cleaned_data['school_other'] = ''
            self.cleaned_data['school_class'] = ''
        return self.cleaned_data


class BirthNumberMixin:
    def clean_birth_num(self):
        qs = SubjectRegistrationParticipant.objects.filter(
            registration__subject_id=self.subject.id,
            registration__canceled=None,
            birth_num=self.cleaned_data['birth_num'],
        )
        if self.instance.pk:
            qs = qs.exclude(pk=self.instance.pk)
        if qs.exists():
            raise ValidationError(_('Participant with this birth number has already been registered.'))
        return self.cleaned_data['birth_num']


class RegistrationParticipantForm(FormMixin, BirthNumberMixin, SchoolMixin, forms.ModelForm):
    x_group = 'age_group'

    school = forms.ChoiceField(
        label=_('School'),
        choices=lambda: (
            [('', '---------')] + [
                (school.id, school) for school in School.objects.all()
            ] + [('other', _('other school'))]
        ),
        required=False,
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # choices for age group
        self.fields['age_group'].widget.choices.queryset = self.subject.age_groups
        if self.subject.age_groups.count() == 1:
            self.fields['age_group'].initial = self.subject.age_groups.first()

        # initial school
        if School.objects.count() == 0:
            self.fields['school'].initial = 'other'

        # prepare data
        prefix = self.prefix + '-'
        data = {
            key[len(prefix):]: value
            for key, value in kwargs.get('data', {}).items()
            if key.startswith(prefix)
        }

        # dynamically required fields
        try:
            age = get_age(get_birth_date(data['birth_num']))
        except Exception:
            age = None

        if age is not None and age < 18:
            self.hide_parents = False
            self.fields['has_parent1'].required = True
        else:
            self.hide_parents = True
            self.fields['phone'].required = True
            self.fields['email'].required = True

        has_parent = ['has_parent1' in data, 'has_parent2' in data]
        for n in range(2):
            if has_parent[n]:
                for field in ['first_name', 'last_name', 'street', 'city', 'postal_code', 'phone', 'email']:
                    self.fields['parent{}_{}'.format(n + 1, field)].required = True

        class ParticipantSelectForm(FormMixin, forms.Form):
            participant = forms.ChoiceField(
                label=_('Choose'),
                choices=[('new', _('new participant'))] + list((p.id, p) for p in self.user_participants),
                initial='new',
                widget=RadioSelectBootstrap(),
            )

        kwargs['prefix'] = self.prefix + '-participant_select'
        self.participant_select_form = ParticipantSelectForm(**kwargs)

        class ParentSelectForm(FormMixin, forms.Form):
            parent = forms.ChoiceField(
                label=_('Choose'),
                choices=[('new', _('new parent'))] + list((p.id, p) for p in self.user_parents),
                initial='new',
                widget=RadioSelectBootstrap(),
            )

        kwargs['prefix'] = self.prefix + '-parent1_select'
        self.parent1_select_form = ParentSelectForm(**kwargs)
        kwargs['prefix'] = self.prefix + '-parent2_select'
        self.parent2_select_form = ParentSelectForm(**kwargs)

        kwargs['prefix'] = self.prefix + '-questions'
        self.questions_form = type(str('QuestionsForm'), (FormMixin, forms.Form), dict(
            (q.name, q.get_field()) for q in self.subject.all_questions
        ))(**kwargs)

    @cached_property
    def required_forms(self):
        return [
            super(RegistrationParticipantForm, self),
            self.participant_select_form,
            self.questions_form,
            self.parent1_select_form,
            self.parent2_select_form,
        ]

    def is_valid(self):
        # validate all required forms
        results = [form.is_valid() for form in self.required_forms]
        return all(results)

    @cached_property
    def errors(self):
        return {
            (form.prefix + '-' + field_name if form.prefix else field_name): error
            for form in self.required_forms
            for field_name, error in form.errors.items()
        }

    def save(self, commit=True):
        with transaction.atomic():
            return self._save(commit)

    def _save(self, commit):
        # set answers
        self.instance.answers = dumps(self.questions_form.cleaned_data)

        # create
        super().save(commit)

        # save / update participant
        if self.participant_select_form.cleaned_data['participant'] == 'new':
            try:
                participant = Participant.objects.get(
                    user=self.instance.registration.user,
                    birth_num=self.instance.birth_num,
                )
            except Participant.DoesNotExist:
                participant = Participant()
                participant.user = self.instance.registration.user
        else:
            participant = Participant.objects.get(id=self.participant_select_form.cleaned_data['participant'])
        for attr in ['first_name', 'last_name', 'birth_num', 'age_group', 'street', 'city', 'postal_code',
                     'citizenship', 'phone', 'email', 'school', 'school_other', 'school_class', 'health']:
            setattr(participant, attr, getattr(self.instance, attr))
        participant.save()

        # save / update first parent
        if self.instance.parent1:
            if self.parent1_select_form.cleaned_data['parent'] == 'new':
                parent = Parent()
                parent.user = self.instance.registration.user
            else:
                parent = Parent.objects.get(id=self.parent1_select_form.cleaned_data['parent'])
            for attr in ['first_name', 'last_name', 'street', 'city', 'postal_code', 'phone', 'email']:
                setattr(parent, attr, getattr(self.instance.parent1, attr))
            parent.save()

        # save / update second parent
        if self.instance.parent2:
            if self.parent2_select_form.cleaned_data['parent'] == 'new':
                parent = Parent()
                parent.user = self.instance.registration.user
            else:
                parent = Parent.objects.get(id=self.parent2_select_form.cleaned_data['parent'])
            for attr in ['first_name', 'last_name', 'street', 'city', 'postal_code', 'phone', 'email']:
                setattr(parent, attr, getattr(self.instance.parent2, attr))
            parent.save()
        return self.instance


class RegistrationGroupForm(FormMixin, SchoolMixin, forms.ModelForm):
    x_group = 'target_group'

    school = forms.ChoiceField(
        label=_('School'),
        choices=lambda: (
            [('', '---------')] + [
                (school.id, school) for school in School.objects.all()
            ] + [('other', _('other school'))]
        ),
        required=False,
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # choices for age group
        self.fields['target_group'].widget.choices.queryset = self.subject.target_groups
        if self.subject.target_groups.count() == 1:
            self.fields['target_group'].initial = self.subject.target_groups.first()

        # initial school
        self.fields['school'].initial = 'other' if School.objects.count() == 0 else None

        kwargs['prefix'] = self.prefix + '-questions'
        self.questions_form = type(str('QuestionsForm'), (FormMixin, forms.Form), dict(
            (q.name, q.get_field()) for q in self.subject.all_questions
        ))(**kwargs)

    @cached_property
    def required_forms(self):
        return [
            super(RegistrationGroupForm, self),
            self.questions_form,
        ]

    def is_valid(self):
        # validate all required forms
        results = [form.is_valid() for form in self.required_forms]
        return all(results)

    @cached_property
    def errors(self):
        return {
            (form.prefix + '-' + field_name if form.prefix else field_name): error
            for form in self.required_forms
            for field_name, error in form.errors.items()
        }

    def save(self, commit=True):
        # set answers
        self.instance.answers = dumps(self.questions_form.cleaned_data)

        # create
        return super().save(commit)


class RegistrationGroupMemberForm(FormMixin, forms.ModelForm):
    pass


class RegistrationBillingInfoForm(FormMixin, forms.ModelForm):
    use_required_attribute = False

    class Meta:
        model = SubjectRegistrationBillingInfo
        exclude = ['registration']


class RegistrationForm(FormMixin, forms.ModelForm):
    def __init__(self, subject, user, **kwargs):
        super(RegistrationForm, self).__init__(**kwargs)
        self.user = user
        self.instance.subject = subject

        # choices for subject_variant
        if self.instance.subject.all_variants:
            self.fields['subject_variant'].widget.choices.queryset = self.instance.subject.variants
        else:
            del self.fields['subject_variant']

        # sub forms
        if subject.registration_type_participants:
            registered_birth_nums = list(SubjectRegistrationParticipant.objects.filter(
                registration_id__in=self.instance.subject.active_registrations.values('id')
            ).values_list('birth_num', flat=True))

            participant_form = type(
                RegistrationParticipantForm.__name__,
                (RegistrationParticipantForm, ),
                {
                    'subject': subject,
                    'user_participants': user.leprikon_participants.exclude(
                        birth_num__in=registered_birth_nums,
                    ),
                    'user_parents': list(user.leprikon_parents.all())
                }
            )

            self.participants_formset = inlineformset_factory(
                SubjectRegistration,
                SubjectRegistrationParticipant,
                form=participant_form,
                fk_name='registration',
                exclude=[],
                extra=subject.max_participants_count,
                min_num=subject.min_participants_count,
                max_num=subject.max_participants_count,
                validate_min=True,
                validate_max=True,
            )(kwargs.get('data'), instance=self.instance)

        elif subject.registration_type_groups:
            group_form = type(
                RegistrationGroupForm.__name__,
                (RegistrationGroupForm, ),
                {'subject': subject}
            )

            self.group_formset = inlineformset_factory(
                SubjectRegistration,
                SubjectRegistrationGroup,
                form=group_form,
                fk_name='registration',
                exclude=[],
                extra=1,
                min_num=1,
                max_num=1,
                validate_min=True,
                validate_max=True,
            )(kwargs.get('data'), instance=self.instance)

            self.group_members_formset = inlineformset_factory(
                SubjectRegistration,
                SubjectRegistrationGroupMember,
                form=RegistrationGroupMemberForm,
                fk_name='registration',
                exclude=[],
                extra=subject.max_group_members_count,
                min_num=subject.min_group_members_count,
                max_num=subject.max_group_members_count,
                validate_min=True,
                validate_max=True,
            )(kwargs.get('data'), instance=self.instance)

        del kwargs['instance']

        self.agreement_forms = []
        for agreement in self.instance.subject.all_registration_agreements:
            kwargs['prefix'] = 'agreement_%s' % agreement.id
            form_attributes = {
                'agreement': agreement,
                'options': {},
            }
            for option in agreement.all_options:
                option_id = 'option_%s' % option.id
                form_attributes[option_id] = AgreementBooleanField(
                    label=option.option,
                    allow_disagree=not option.required,
                )
                form_attributes['options'][option_id] = option
            AgreementForm = type(str('AgreementForm'), (FormMixin, forms.Form), form_attributes)
            self.agreement_forms.append(AgreementForm(**kwargs))

        self.user_billing_info = list(self.user.leprikon_billing_info.all())

        class BillingInfoSelectForm(FormMixin, forms.Form):
            billing_info = forms.ChoiceField(
                label=_('Choose'),
                choices=[
                    ('none', _('no billing information')),
                ] + list((bi.id, bi) for bi in self.user_billing_info) + [
                    ('new', _('new billing information')),
                ],
                initial='none',
                widget=RadioSelectBootstrap(),
            )

        kwargs['prefix'] = 'billing_info_select'
        self.billing_info_select_form = BillingInfoSelectForm(**kwargs)

        kwargs['prefix'] = 'billing_info'
        self.billing_info_form = RegistrationBillingInfoForm(**kwargs)

    @cached_property
    def required_forms(self):
        return [
            super(RegistrationForm, self),
            self.billing_info_select_form,
        ] + self.agreement_forms

    @cached_property
    def all_forms(self):
        all_forms = self.required_forms
        if self.instance.subject.registration_type_participants:
            all_forms += self.participants_formset.forms
        elif self.instance.subject.registration_type_groups:
            all_forms += self.group_formset.forms + self.group_members_formset.forms
        if (
            not self.billing_info_select_form.is_valid() or
            self.billing_info_select_form.cleaned_data['billing_info'] != 'none'
        ):
            all_forms.append(self.billing_info_form)
        return all_forms

    def is_valid(self):
        # validate all required forms
        results = [form.is_valid() for form in self.required_forms]
        if self.instance.subject.registration_type_participants:
            results.append(self.participants_formset.is_valid())
        elif self.instance.subject.registration_type_groups:
            results += [
                self.group_formset.is_valid(),
                self.group_members_formset.is_valid(),
            ]
        if (
            not self.billing_info_select_form.is_valid() or
            self.billing_info_select_form.cleaned_data['billing_info'] != 'none'
        ):
            results.append(self.billing_info_form.is_valid())
        return all(results)

    @cached_property
    def errors(self):
        return {
            (form.prefix + field_name if form.prefix else field_name): error
            for form in self.all_forms
            for field_name, error in form.errors.items()
        }

    def save(self, commit=True):
        with transaction.atomic():
            return self._save(commit)

    def _save(self, commit):
        # set user
        self.instance.user = self.user

        # set price
        self.instance.price = (
            self.instance.subject_variant.price
            if self.instance.subject_variant
            else self.instance.subject.price
        )

        # create
        super().save(True)

        if self.instance.subject.registration_type_participants:
            # save participants
            self.participants_formset.user = self.instance.user
            self.participants_formset.save()

        elif self.instance.subject.registration_type_groups:
            # save group
            self.group_formset.save()

            # save group members
            self.group_members_formset.save()

        # save additional questions
        self.instance.questions.set(self.instance.subject.all_questions)

        # save legal agreements
        self.instance.agreements.set(self.instance.subject.all_registration_agreements)

        # save agreement options
        for agreement_form in self.agreement_forms:
            for option_id, checked in agreement_form.cleaned_data.items():
                if checked:
                    self.instance.agreement_options.add(agreement_form.options[option_id])

        if self.billing_info_select_form.cleaned_data['billing_info'] != 'none':

            # save registration billing info
            registration_billing_info = self.billing_info_form.save(commit=False)
            registration_billing_info.registration = self.instance
            registration_billing_info.save()

            # save / update user billing info
            if self.billing_info_select_form.cleaned_data['billing_info'] == 'new':
                billing_info = BillingInfo()
                billing_info.user = self.instance.user
            else:
                billing_info = BillingInfo.objects.get(id=self.billing_info_select_form.cleaned_data['billing_info'])
            for attr in [
                'name', 'street', 'city', 'postal_code', 'company_num', 'vat_number',
                'contact_person', 'phone', 'email', 'employee',
            ]:
                setattr(billing_info, attr, getattr(registration_billing_info, attr))
            billing_info.save()

        self.instance.generate_variable_symbol_and_slug()
        self.instance.send_mail()
        return self.instance


class CourseRegistrationForm(RegistrationForm):

    class Meta:
        model = CourseRegistration
        exclude = ('subject', 'user', 'cancel_request', 'canceled')


class EventRegistrationForm(RegistrationForm):

    class Meta:
        model = EventRegistration
        exclude = ('subject', 'user', 'cancel_request', 'canceled')


class RegistrationAdminForm(forms.ModelForm):

    def __init__(self, data=None, *args, **kwargs):
        super().__init__(data, *args, **kwargs)

        if 'subject_variant' in self.fields:
            self.fields['subject_variant'].widget.choices.queryset = self.subject.variants

        # choices for agreement options
        instance = kwargs.get('instance')
        self.fields['agreement_options'].widget.choices = tuple(
            (agreement.name, tuple((option.id, option.name) for option in agreement.all_options))
            for agreement in (instance.all_agreements if instance else self.subject.all_registration_agreements)
        )


class SchoolAdminMixin:
    def _add_error_required(self, field_name):
        self.add_error(field_name, forms.ValidationError(
            self.fields[field_name].error_messages['required'],
            code='required',
        ))

    def clean(self):
        # check required school and class
        x_group = self.cleaned_data.get(self.x_group)
        if x_group and x_group.require_school:
            # require school
            if self.cleaned_data.get('school'):
                self.cleaned_data['school_other'] = ''
            elif not self.cleaned_data.get('school_other'):
                self._add_error_required('school_other')
            # require school class
            if not self.cleaned_data.get('school_class'):
                self._add_error_required('school_class')
        else:
            # delete school and class
            self.cleaned_data['school'] = None
            self.cleaned_data['school_other'] = ''
            self.cleaned_data['school_class'] = ''
        return self.cleaned_data


class RegistrationParticipantAdminForm(BirthNumberMixin, SchoolAdminMixin, forms.ModelForm):
    x_group = 'age_group'

    def __init__(self, data=None, *args, **kwargs):
        super().__init__(data, *args, **kwargs)
        instance = kwargs.get('instance')

        self.fields['age_group'].widget.choices.queryset = self.subject.age_groups
        if self.subject.age_groups.count() == 1:
            self.fields['age_group'].initial = self.subject.age_groups.first()

        questions = self.obj.all_questions if self.obj else self.subject.all_questions
        answers = instance.get_answers() if instance else {}
        for q in questions:
            self.fields['q_' + q.name].initial = answers.get(q.name)

        created_date = self.obj.created.date() if self.obj else date.today()
        try:
            age = get_age(get_birth_date(data['birth_num']), created_date)
        except Exception:
            age = get_age(get_birth_date(instance.birth_num), created_date) if instance else None
        if age is not None:
            if age < 18:
                self.fields['has_parent1'].required = True
            else:
                self.fields['phone'].required = True
                self.fields['email'].required = True

        try:
            has_parent = ['has_parent1' in data, 'has_parent2' in data]
        except TypeError:
            has_parent = [instance.has_parent1, instance.has_parent2] if instance else [False, False]
        for n in range(2):
            if has_parent[n]:
                for field in ['first_name', 'last_name', 'street', 'city', 'postal_code', 'phone', 'email']:
                    self.fields['parent{}_{}'.format(n + 1, field)].required = True

    def save(self, commit=True):
        self.instance.answers = dumps({
            q.name: self.cleaned_data['q_' + q.name]
            for q in (self.obj.all_questions if self.obj else self.subject.all_questions)
        })
        return super().save(commit)


class RegistrationGroupAdminForm(SchoolAdminMixin, forms.ModelForm):
    x_group = 'target_group'

    def __init__(self, data=None, *args, **kwargs):
        super().__init__(data, *args, **kwargs)
        instance = kwargs.get('instance')

        self.fields['target_group'].widget.choices.queryset = self.subject.target_groups

        questions = self.obj.all_questions if self.obj else self.subject.all_questions
        answers = instance.get_answers() if instance else {}
        for q in questions:
            self.fields['q_' + q.name].initial = answers.get(q.name)

    def save(self, commit=True):
        self.instance.answers = dumps({
            q.name: self.cleaned_data['q_' + q.name]
            for q in (self.obj.all_questions if self.obj else self.subject.all_questions)
        })
        return super().save(commit)



from django.conf.urls import url

from . import views


def l_url(pattern, name):
    return url(pattern, getattr(views, name), name=name)


urlpatterns = [
    l_url(r'^participants/(?P<subject_id>[0-9]+)/$', 'participants'),
]



from string import ascii_letters, digits

from django.contrib import admin
from django.db import transaction
from django.db.models import Count
from django.db.utils import IntegrityError
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.crypto import get_random_string
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _

from ..forms.registrationlink import RegistrationLinkAdminForm
from ..models.courses import CourseRegistration
from ..models.events import EventRegistration
from ..models.registrationlink import RegistrationLink
from ..models.subjects import SubjectType
from .export import AdminExportMixin
from .filters import SchoolYearListFilter, SubjectTypeListFilter


@admin.register(RegistrationLink)
class RegistrationLinkAdmin(AdminExportMixin, admin.ModelAdmin):
    _registration_models = {
        SubjectType.COURSE: CourseRegistration,
        SubjectType.EVENT: EventRegistration,
    }
    list_display = (
        'id', 'name', 'get_link', 'subject_type', 'reg_from', 'reg_to', 'get_registrations_link',
    )
    list_filter = (
        ('school_year', SchoolYearListFilter),
        'subject_type__subject_type',
        ('subject_type', SubjectTypeListFilter),
    )
    filter_horizontal = ('subjects',)

    def changeform_view(self, request, object_id=None, form_url='', extra_context=None):
        if not object_id and request.method == 'POST' and len(request.POST) == 3:
            return HttpResponseRedirect('{}?subject_type={}'.format(
                request.path,
                request.POST.get('subject_type', ''),
            ))
        else:
            return super().changeform_view(request, object_id, form_url, extra_context)

    def get_form(self, request, obj, **kwargs):
        # set school year
        if obj:
            request.school_year = obj.school_year

        # get subject type
        try:
            # first try request.POST (user may want to change subject type)
            request.subject_type = SubjectType.objects.get(id=int(request.POST.get('subject_type')))
        except (SubjectType.DoesNotExist, TypeError, ValueError):
            if obj:
                # use subject type from object
                request.subject_type = obj.subject_type
            else:
                # try to get subject type from request.GET
                try:
                    request.subject_type = SubjectType.objects.get(
                        id=int(request.GET.get('subject_type')),
                    )
                except (SubjectType.DoesNotExist, TypeError, ValueError):
                    request.subject_type = None

        if request.subject_type:
            kwargs['form'] = type(
                RegistrationLinkAdminForm.__name__,
                (RegistrationLinkAdminForm, ),
                {
                    'school_year': request.school_year,
                    'subject_type': request.subject_type,
                },
            )
        else:
            kwargs['fields'] = ['subject_type']

        return super().get_form(request, obj, **kwargs)

    def save_form(self, request, form, change):
        obj = super().save_form(request, form, change)
        obj.school_year = request.school_year
        return obj

    def save_model(self, request, obj, form, change):
        if change:
            obj.save()
        else:
            obj.school_year = request.school_year
            while not obj.id:
                try:
                    with transaction.atomic():
                        obj.slug = get_random_string(64, ascii_letters + digits)
                        obj.save()
                except IntegrityError:
                    pass

    def get_link(self, obj):
        return '<a href="{url}" title="{title}" target="_blank">{url}</a>'.format(
            url=obj.link,
            title=_('registration link'),
        )
    get_link.short_description = _('registration link')
    get_link.allow_tags = True

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request).select_related('subject_type')
            .annotate(registrations_count=Count('registrations'))
        )

    def get_registrations_link(self, obj):
        registration_model = self._registration_models[obj.subject_type.subject_type]
        return format_html(
            '<a href="{url}">{count}</a>',
            url=reverse('admin:{}_{}_changelist'.format(
                registration_model._meta.app_label,
                registration_model._meta.model_name,
            )) + '?registration_link__id__exact={}'.format(obj.id),
            count=obj.registrations_count,
        )
    get_registrations_link.short_description = _('registrations')
    get_registrations_link.allow_tags = True



from django.contrib import admin

from ..models.agreements import Agreement, AgreementOption
from .filters import ActiveListFilter


class AgreementOptionInlineAdmin(admin.TabularInline):
    model = AgreementOption
    extra = 0
    min_num = 1


@admin.register(Agreement)
class AgreementAdmin(admin.ModelAdmin):
    inlines = (AgreementOptionInlineAdmin,)
    list_display = ('name', 'order')
    list_editable = ('order',)
    list_filter = (ActiveListFilter,)

    def has_delete_permission(self, request, obj=None):
        if obj and obj.registrations.exists():
            return False
        else:
            return super().has_delete_permission(request, obj)

    def get_actions(self, request):
        actions = super().get_actions(request)
        if 'delete_selected' in actions:
            del(actions['delete_selected'])
        return actions



from django import forms

from ..models.courses import CourseDiscount


class CourseDiscountAdminForm(forms.ModelForm):

    class Meta:
        model = CourseDiscount
        fields = ['registration', 'period', 'accounted', 'amount', 'explanation']

    def __init__(self, *args, **kwargs):
        super(CourseDiscountAdminForm, self).__init__(*args, **kwargs)
        if 'period' in self.fields:
            if self.data:
                registration_id = int(self.data['registration'])
            elif self.initial:
                registration_id = int(self.initial['registration'])
            else:
                registration_id = self.instance.registration_id
            self.fields['period'].widget.choices.queryset = self.fields['period'].widget.choices.queryset.filter(
                school_year_division__courses__registrations__id=registration_id,
            )



# Generated by Django 1.9.13 on 2018-01-05 12:45

from django.db import migrations


def set_is_city_part(apps, schema_editor):
    CourseDiscount = apps.get_model('leprikon', 'CourseDiscount')
    SchoolYearPeriod = apps.get_model('leprikon', 'SchoolYearPeriod')
    for discount in CourseDiscount.objects.all():
        periods = list(SchoolYearPeriod.objects.filter(
            courses__registrations__id=discount.registration_id,
            start__lte=discount.period.end,
            end__gte=discount.period.start,
        ))
        if discount.period not in periods:
            if periods:
                discount.period = periods[0]
                discount.save()
                print('Fixed period for discount %d' % discount.id)
            else:
                print('Failed to fix period for discount %d' % discount.id)


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0005_subjectregistration_variable_symbol'),
    ]

    operations = [
        migrations.RunPython(set_is_city_part),
    ]



from django import forms
from django.conf.urls import url as urls_url
from django.contrib import admin
from django.contrib.auth import get_user_model, login
from django.contrib.auth.admin import UserAdmin as _UserAdmin
from django.contrib.auth.decorators import permission_required
from django.contrib.messages import ERROR
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from ..forms.user import UserAdminChangeForm, UserAdminCreateForm
from ..utils import merge_users
from .messages import SendMessageAdminMixin

User = get_user_model()
admin.site.unregister(User)


@admin.register(User)
class UserAdmin(SendMessageAdminMixin, _UserAdmin):
    actions = ('merge',)
    add_form = UserAdminCreateForm
    add_fieldsets = (
        (None, {
            'classes': ('wide',),
            'fields': ('username', 'first_name', 'last_name', 'email'),
        }),
    )
    form = UserAdminChangeForm

    def merge(self, request, queryset):
        class MergeForm(forms.Form):
            target = forms.ModelChoiceField(
                label=_('Target user'),
                help_text=_('All information will be merged into selected account.'),
                queryset=queryset,
            )
        if request.POST.get('post', 'no') == 'yes':
            form = MergeForm(request.POST)
            if form.is_valid():
                target = form.cleaned_data['target']
                for user in queryset.all():
                    if user != target:
                        try:
                            merge_users(user, target)
                            self.message_user(request, _('User {} was merged into user {}.').format(user, target))
                        except Exception:
                            self.message_user(
                                request, _('Can not merge user {} into user {}.').format(user, target),
                                level=ERROR
                            )
                return
        else:
            form = MergeForm()
        return render(request, 'leprikon/admin/merge.html', {
            'title': _('Select target user for merge'),
            'question': _('Are you sure you want to merge selected users into one? '
                          'All participants, parents, registrations and other related information '
                          'will be added to the target user account and the remaining users will be deleted.'),
            'queryset': queryset,
            'objects_title': _('Users'),
            'form_title': _('Select target account for merge'),
            'opts': self.model._meta,
            'form': form,
            'action_checkbox_name': admin.helpers.ACTION_CHECKBOX_NAME,
        })
    merge.short_description = _('Merge selected user accounts')

    def get_list_display(self, request):
        list_display = ['id'] + list(super(UserAdmin, self).get_list_display(request))
        if request.user.is_superuser:
            list_display += ['login_as_link_superuser']
        elif request.user.has_perm('auth.change_user'):
            list_display += ['login_as_link']
        return list_display

    def get_search_fields(self, request):
        return list(super(UserAdmin, self).get_search_fields(request)) + [
            'leprikon_parents__first_name',
            'leprikon_parents__last_name',
            'leprikon_parents__email',
            'leprikon_participants__first_name',
            'leprikon_participants__last_name',
            'leprikon_participants__email',
        ]

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = list(super(UserAdmin, self).get_readonly_fields(request, obj)) + [
            'last_login',
            'date_joined',
        ]
        if not request.user.is_superuser:
            readonly_fields += ['is_staff', 'is_superuser', 'groups', 'user_permissions']
        return readonly_fields

    def get_urls(self):
        urls = super(UserAdmin, self).get_urls()
        login_as_view = self.admin_site.admin_view(permission_required('auth.change_user')(self.login_as))
        return [
            urls_url(r'(?P<user_id>\d+)/login-as/$', login_as_view, name='auth_user_login_as'),
        ] + urls

    def login_as(self, request, user_id):
        if request.user.is_superuser:
            user = get_object_or_404(get_user_model(), id=user_id)
        else:
            user = get_object_or_404(get_user_model(), id=user_id, is_staff=False, is_superuser=False)
        user.backend = 'django.contrib.auth.backends.ModelBackend'
        login(request, user)
        return HttpResponseRedirect(reverse('leprikon:summary'))

    def login_as_link(self, obj):
        return '<a href="{url}">{text}</a>'.format(
            url=reverse('admin:auth_user_login_as', args=[obj.id]),
            text=_('login')
        ) if not obj.is_staff and not obj.is_superuser else ''
    login_as_link.allow_tags = True
    login_as_link.short_description = _('login')

    def login_as_link_superuser(self, obj):
        return '<a href="{url}">{text}</a>'.format(
            url=reverse('admin:auth_user_login_as', args=[obj.id]),
            text=_('login')
        )
    login_as_link_superuser.allow_tags = True
    login_as_link_superuser.short_description = _('login')

    def get_message_recipients(self, request, queryset):
        return queryset.all()



from . import (  # noqa
    account, agegroup, agreements, citizenship, courses, department, events,
    fields, journals, leprikonsite, messages, organizations, place, printsetup,
    question, registrationlink, roles, school, schoolyear, startend, subjects,
    targetgroup, timesheets, useragreement, utils,
)



from django.utils.translation import ugettext_lazy as _

PRICE_DECIMAL_PLACES = 2
PRICE_MAX_DIGITS = 10

LEPRIKON_SHOW_SUBJECT_CODE = False

LEPRIKON_COUNTRY = 'CZ'

LEPRIKON_CONTACT_TYPES = (
    ('email', _('email')),
    ('phone', _('phone')),
    ('url', _('url')),
)

LEPRIKON_COURSE_TEMPLATES = (
    ('default', _('Default')),
)

LEPRIKON_EVENT_TEMPLATES = (
    ('default', _('Default')),
)

LEPRIKON_LEADER_TEMPLATES = (
    ('default', _('Default')),
)

LEPRIKON_COURSELIST_TEMPLATES = (
    ('default', _('Default')),
    ('grouped', _('Grouped by course groups')),
)

LEPRIKON_EVENTLIST_TEMPLATES = (
    ('default', _('Default')),
    ('grouped', _('Grouped by event groups')),
)

LEPRIKON_LEADERLIST_TEMPLATES = (
    ('default', _('Default')),
)

LEPRIKON_COLOR_POSITIVE = LEPRIKON_COLOR_PAID = '#0a0'
LEPRIKON_COLOR_NEGATIVE = LEPRIKON_COLOR_NOTPAID = '#e60'
LEPRIKON_COLOR_ZERO = LEPRIKON_COLOR_OVERPAID = '#00a'
LEPRIKON_COLOR_OVERDUE = '#e00'

LEPRIKON_QUESTION_FIELDS = {
    'boolean': {
        'name': _('boolean field'),
        'class': 'django.forms.BooleanField',
    },
    'char': {
        'name': _('char field'),
        'class': 'django.forms.CharField',
    },
    'text': {
        'name': _('text field'),
        'class': 'leprikon.forms.fields.TextField',
    },
    'choice': {
        'name': _('choice field'),
        'class': 'django.forms.ChoiceField',
    },
    'date': {
        'name': _('date field'),
        'class': 'django.forms.DateField',
    },
}

LEPRIKON_PARAM_BACK = 'zpet'

LEPRIKON_URL = None

LEPRIKON_URL_SUMMARY = 'prehled'
LEPRIKON_URL_LEADER = 'pro-vedouci'
LEPRIKON_URL_CREATE_ACCOUNT = 'novy-uzivatel'
LEPRIKON_URL_COURSES = 'kurzy'
LEPRIKON_URL_EVENTS = 'akce'
LEPRIKON_URL_LEADERS = 'vedouci'
LEPRIKON_URL_MESSAGES = 'zpravy'
LEPRIKON_URL_TIMESHEETS = 'vykazy'
LEPRIKON_URL_SUBMIT = 'odevzdat'
LEPRIKON_URL_ENTRY = 'zaznam'
LEPRIKON_URL_PARENT = 'zastupce'
LEPRIKON_URL_PARTICIPANT = 'ucastnik'
LEPRIKON_URL_PARTICIPANTS = 'ucastnici'
LEPRIKON_URL_BILLING_INFO = 'fakturacni-udaje'
LEPRIKON_URL_USER = 'uzivatel'
LEPRIKON_URL_AGREEMENT = 'souhlas'
LEPRIKON_URL_PASSWORD = 'heslo'
LEPRIKON_URL_PASSWORD_RESET = 'reset-hesla'
LEPRIKON_URL_LOGIN = 'prihlasit'
LEPRIKON_URL_LOGOUT = 'odhlasit'
LEPRIKON_URL_SCHOOL_YEAR = 'skolni-rok'
LEPRIKON_URL_SUPPORT = 'pomoc'
LEPRIKON_URL_EDIT = 'upravit'
LEPRIKON_URL_JOURNAL = 'denik'
LEPRIKON_URL_JOURNALENTRY = 'zaznam-deniku'
LEPRIKON_URL_MINE = 'moje'
LEPRIKON_URL_ALTERNATING = 'zastup'
LEPRIKON_URL_REPORTS = 'reporty'
LEPRIKON_URL_PAYMENT_REQUEST = 'vyzva-k-platbe'
LEPRIKON_URL_PAYMENTS = 'platby'
LEPRIKON_URL_PAYMENTS_STATUS = 'stav-plateb'
LEPRIKON_URL_DEBTORS = 'dluznici'
LEPRIKON_URL_ADD = 'pridat'
LEPRIKON_URL_DELETE = 'smazat'
LEPRIKON_URL_CANCEL = 'zrusit'
LEPRIKON_URL_REGISTRATION = 'prihlaska'
LEPRIKON_URL_REGISTRATIONS = 'prihlasky'
LEPRIKON_URL_REGISTRATION_LINK = 'prihlasovaci-odkaz'
LEPRIKON_URL_TERMS_CONDITIONS = 'obchodni-podminky'
LEPRIKON_URL_STATS = 'statistiky'
LEPRIKON_URL_EMAIL = 'email'

LEPRIKON_MENU_ADD_LOGOUT = True

LEPRIKON_SITE_CACHE_MAX_AGE = 5

# expression to create variable symbol (subject.code + last two digits of year + last four digits of id)
LEPRIKON_VARIABLE_SYMBOL_EXPRESSION = 'reg.subject.code * 1000000 + (reg.created.year % 100) * 10000 + (reg.id % 10000)'

SESSION_STAFF_COOKIE_AGE = 60 * 60 * 10
